#ifndef _OWLPS_POSITIONING_ACCESSPOINT_HH_
#define _OWLPS_POSITIONING_ACCESSPOINT_HH_

#include "wifidevice.hh"
#include "point3d.hh"
#include "posutil.hh"

#define AP_DEFAULT_CHANNEL 6
#define AP_DEFAULT_ANTENNA_GAIN 5

/// Represents a capture \link WifiDevice Wi-Fi device\endlink
class AccessPoint: public WifiDevice
{
protected:
  Point3D coordinates ;
  unsigned long frequency ; ///< Frequency (channel) in Hz
  float friis_index ; ///< Friis index used by FBCM

public:
  /**
   * Special parameters:
   * - \em _antenna_gain Antenna gain in dBi.
   * - \em _trx_power Transmit power in dBm.
   * - \em channel Wi-Fi channel the AP is listening to (integer
   * between 1 and 14). It will be converted to a frequency in Hz.
   */
  AccessPoint(const Point3D &_coordinates = Point3D(),
              const std::string &_ip_addr = "",
              const std::string &_mac_addr = "",
              const float _antenna_gain = AP_DEFAULT_ANTENNA_GAIN,
              const float _trx_power = WIFIDEVICE_DEFAULT_TRX_POWER,
              const unsigned int &channel = AP_DEFAULT_CHANNEL):
    WifiDevice(_ip_addr, _mac_addr, _antenna_gain, _trx_power),
    coordinates(_coordinates),
    frequency(PosUtil::wifi_channel_to_hz(channel)),
    friis_index(0) {}

  AccessPoint(const WifiDevice &source,
              const Point3D &_coordinates,
              const unsigned int channel = AP_DEFAULT_CHANNEL):
    WifiDevice(source), coordinates(_coordinates),
    frequency(PosUtil::wifi_channel_to_hz(channel)),
    friis_index(0) {}

  AccessPoint(const WifiDevice &source):
    WifiDevice(source), coordinates(Point3D()), frequency(0),
    friis_index(0) {}

  AccessPoint(const AccessPoint &source):
    WifiDevice(source), coordinates(source.coordinates),
    frequency(source.frequency), friis_index(0) {}

  ~AccessPoint(void) {}

  /** @name Read accessors */
  //@{
  const Point3D& get_coordinates(void) const ;
  unsigned long get_frequency(void) const ;
  float get_friis_index(void) const ;
  //@}

  /** @name Write accessors */
  //@{
  void set_coordinates(const Point3D &_coordinates) ;
  void set_channel(const unsigned int channel) ;
  void set_frequency(const unsigned long _frequency) ;
  void set_friis_index(const float _friis_index) ;
  //@}

  /** @name Operators */
  //@{
  const AccessPoint& operator=(const AccessPoint &source) ;
  bool operator==(const AccessPoint &source) const ;
  bool operator!=(const AccessPoint &source) const ;
  //@}

  /// Displays an AccessPoint
  friend std::ostream &operator<<(std::ostream &os, const AccessPoint &ap) ;
} ;



/* *** Read accessors *** */


inline const Point3D& AccessPoint::get_coordinates() const
{
  return coordinates ;
}


inline unsigned long AccessPoint::get_frequency() const
{
  return frequency ;
}


inline float AccessPoint::get_friis_index() const
{
  return friis_index ;
}



/* *** Write accessors *** */


inline void AccessPoint::set_coordinates(const Point3D &_coordinates)
{
  coordinates = _coordinates ;
}


/**
 * @param channel A Wi-Fi channel (integer between 1 and 14). It will
 * be converted into a frequency in Hz and affected to #frequency. If
 * it is not a Wi-Fi channel number, 0 will be affected to #frequency.
 */
inline void AccessPoint::set_channel(const unsigned int channel)
{
  set_frequency(PosUtil::wifi_channel_to_hz(channel)) ;
}


/**
 * Sets #frequency with the value of the parameter \em _frequency. No
 * control is done, so you should pass a correct value.
 *
 * Note that set_channel() is more secure because a wrong channel
 * number will cause #frequency to be nullified.
 */
inline void AccessPoint::set_frequency(const unsigned long _frequency)
{
  frequency = _frequency ;
}


inline void AccessPoint::set_friis_index(const float _friis_index)
{
  friis_index = _friis_index ;
}



/* *** Operators *** */


inline bool AccessPoint::operator!=(const AccessPoint &source) const
{
  return !(*this == source) ;
}



#endif // _OWLPS_POSITIONING_ACCESSPOINT_HH_
