#ifndef _OWLPS_POSITIONING_REFERENCEPOINT_HH_
#define _OWLPS_POSITIONING_REFERENCEPOINT_HH_

class AccessPoint ;
class CalibrationRequest ;
class Request ;
class Measurement ;

#include "point3d.hh"

#include <vector>
#include <map>
#include <ostream>

/// Represents a reference point in 3-D space
class ReferencePoint: public Point3D
{
protected:
  /// List of CalibrationRequest associated with the ReferencePoint
  /** Note that \em requests is a \em pointer list: only pointers
      are stored, not values. */
  std::vector<CalibrationRequest*> requests ;

  /** @name Operations */
  //@{
  /// Mutually complete two Measurement lists with missing APs
  void complete_with_dummy_measurements(
    std::map<std::string, Measurement> &measurements1,
    std::map<std::string, Measurement> &measurements2) const ;
  /// Compute the distance between two Measurement lists
  float compute_ss_square_distance(
    std::map<std::string, Measurement> &measurements1,
    std::map<std::string, Measurement> &measurements2) const ;
  //@}

public:
  ReferencePoint(const float &_x = 0,
                 const float &_y = 0,
                 const float &_z = 0): Point3D(_x, _y, _z) {}

  ReferencePoint(const Point3D &p): Point3D(p) {}

  ReferencePoint(const ReferencePoint &source):
    Point3D(source), requests(source.requests) {}

  ~ReferencePoint(void) ;

  /** @name Read accessors */
  //@{
  const std::vector<CalibrationRequest*>& get_requests(void) const ;
  std::map<std::string, Measurement>
  get_all_measurements_sorted(void) const ;
  //@}

  /** @name Write accessors */
  //@{
  /// Adds a Request to the \link #requests request list\endlink
  void add_request(const CalibrationRequest *cm) ;
  //@}

  /** @name Operations */
  //@{
  /// Compute the distance between the ReferencePoint and a Request
  float ss_square_distance(const Request &source) const ;
  /// Compute the Friis index for the given AccessPoint
  float friis_index_for_ap(const std::string &ap_mac) const ;
  /// Compute the Friis indexes sum for the given AccessPoint
  float friis_indexes_for_ap(
    const AccessPoint &ap, const double &const_term,
    int &nb_indexes) const ;
  //@}

  /** @name Operators */
  //@{
  const ReferencePoint& operator=(const ReferencePoint &source) ;
  bool operator==(const ReferencePoint &source) const ;
  bool operator!=(const ReferencePoint &source) const ;
  //@}

  /// Displays a ReferencePoint
  friend std::ostream &operator<<(std::ostream &os, const ReferencePoint &rp) ;

  /// Hash a ReferencePoint
  friend std::size_t hash_value(const ReferencePoint &source) ;
} ;



/* *** Read accessors *** */


inline const
std::vector<CalibrationRequest*>& ReferencePoint::get_requests() const
{
  return requests ;
}



/* *** Write accessors *** */


/**
 * @param cm A pointer to the CalibrationRequest to add. If it is
 * NULL, nothing will be done.
 * The memory pointed by this pointer must not be deallocated before
 * the ReferencePoint destruction (do \em not pass a pointer to a local
 * variable!).
 */
inline void ReferencePoint::add_request(const CalibrationRequest *cm)
{
  if (cm != NULL)
    requests.push_back(const_cast<CalibrationRequest*>(cm)) ;
}



/* *** Operators *** */


inline bool ReferencePoint::operator!=(const ReferencePoint &source) const
{
  return !(*this == source) ;
}



#endif // _OWLPS_POSITIONING_REFERENCEPOINT_HH_
