#ifndef _OWLPS_POSITIONING_TIMESTAMP_HH_
#define _OWLPS_POSITIONING_TIMESTAMP_HH_

#include <ctime>
#include <stdint.h> // <cstdint> is not C++ 98 compliant
#include <ostream>

/// Represents a timestamp with a millisecond precision
/**
 * This class boxes <em>struct timespec</em> from ctime (time.h) to
 * integrate time manipulation functions.
 */
class Timestamp
{
protected:
  /// Time data
  struct timespec timestamp ;

  /** @name Internal accessors */
  //@{
  void set(const struct timespec &source) ;
  /// Initialises the Timestamp with a value in milliseconds
  void set(const uint64_t source_ms) ;
  //@}

  /** @name Comparison functions */
  //@{
  bool equals(const struct timespec &source) const ;
  bool equals(const uint64_t source) const ;
  bool less_than(const struct timespec &source) const ;
  bool greater_than(const struct timespec &source) const ;
  //@}

  /** @name Operations */
  //@{
  /// \brief Initialises #timestamp at the current time whith a
  /// nanosecond precision
  bool now_ns(void) ;
  /// Lowers the precision of #timestamp to ms
  void round_to_ms(void) ;
  //@}


public:
  Timestamp(void) ;
  Timestamp(const struct timespec &source) ;
  /// Constructs a Timsestamp from a value in milliseconds
  Timestamp(const uint64_t source) ;
  Timestamp(const Timestamp &source) ;

  ~Timestamp(void) {}

  /// \brief Initialises #timestamp to the current time with a
  /// millisecond precision
  bool now(void) ;

  void clear(void) ;

  /** @name Operators */
  //@{
  const Timestamp& operator=(const Timestamp &source) ;
  const Timestamp& operator=(const struct timespec &source) ;
  const Timestamp& operator=(const uint64_t source) ;
  bool operator==(const Timestamp &source) const ;
  bool operator==(const struct timespec &source) const ;
  bool operator==(const uint64_t source) const ;
  bool operator!=(const Timestamp &source) const ;
  bool operator!=(const struct timespec &source) const ;
  bool operator!=(const uint64_t source) const ;
  bool operator<(const Timestamp &source) const ;
  bool operator<(const struct timespec &source) const ;
  bool operator<(const uint64_t source) const ;
  bool operator>(const Timestamp &source) const ;
  bool operator>(const struct timespec &source) const ;
  bool operator>(const uint64_t source) const ;
  bool operator<=(const Timestamp &source) const ;
  bool operator<=(const struct timespec &source) const ;
  bool operator<=(const uint64_t source) const ;
  bool operator>=(const Timestamp &source) const ;
  bool operator>=(const struct timespec &source) const ;
  bool operator>=(const uint64_t source) const ;
  operator bool(void) const ;
  operator const struct timespec&(void) const ;
  /// Cast to milliseconds
  operator uint64_t(void) const ;
  //@}

  /// Display a Timestamp
  friend std::ostream& operator<<(std::ostream &os, const Timestamp &t) ;

  // Hash a Timestamp
  friend size_t hash_value(const Timestamp &source) ;
} ;



inline void Timestamp::clear(void)
{
  timestamp.tv_sec = 0 ;
  timestamp.tv_nsec = 0 ;
}



/* *** Operators *** */


inline const Timestamp& Timestamp::
operator=(const struct timespec &source)
{
  set(source) ;
  return *this ;
}


inline const Timestamp& Timestamp::operator=(const uint64_t source)
{
  set(source) ;
  return *this ;
}


inline bool Timestamp::operator==(const struct timespec &source) const
{
  return equals(source) ;
}


inline bool Timestamp::operator==(const uint64_t source) const
{
  return equals(source) ;
}


inline bool Timestamp::operator!=(const Timestamp &source) const
{
  return !(*this == source) ;
}


inline bool Timestamp::operator!=(const struct timespec &source) const
{
  return !(*this == source) ;
}


inline bool Timestamp::operator!=(const uint64_t source) const
{
  return !(*this == source) ;
}


inline bool Timestamp::operator<(const Timestamp &source) const
{
  return less_than(source.timestamp) ;
}


inline bool Timestamp::operator<(const struct timespec &source) const
{
  return less_than(source) ;
}


inline bool Timestamp::operator<(const uint64_t source) const
{
  return static_cast<uint64_t>(*this) < source ;
}


inline bool Timestamp::operator>(const Timestamp &source) const
{
  return greater_than(source.timestamp) ;
}


inline bool Timestamp::operator>(const struct timespec &source) const
{
  return greater_than(source) ;
}


inline bool Timestamp::operator>(const uint64_t source) const
{
  return static_cast<uint64_t>(*this) > source ;
}


inline bool Timestamp::operator<=(const Timestamp &source) const
{
  return *this == source || *this < source ;
}


inline bool Timestamp::operator<=(const struct timespec &source) const
{
  return *this == source || *this < source ;
}


inline bool Timestamp::operator<=(const uint64_t source) const
{
  return *this == source || *this < source ;
}


inline bool Timestamp::operator>=(const Timestamp &source) const
{
  return source <= *this ;
}


inline bool Timestamp::operator>=(const struct timespec &source) const
{
  return *this == source || !(*this < source) ;
}


inline bool Timestamp::operator>=(const uint64_t source) const
{
  return *this == source || !(*this < source) ;
}


inline Timestamp::operator bool() const
{
  return static_cast<uint64_t>(*this) > 0 ;
}


inline Timestamp::operator const struct timespec&() const
{
  return timestamp ;
}


inline Timestamp::operator uint64_t(void) const
{
  return
    static_cast<uint64_t>(timestamp.tv_sec) * 1000 +
    static_cast<uint64_t>(timestamp.tv_nsec) / 1000000 ;
}



#endif // _OWLPS_POSITIONING_TIMESTAMP_HH_
